@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2017 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -46,10 +46,6 @@ typedef struct tls_data_struct TLS_DATA;
 
 /**************************************** data structures */
 
-/* non-zero constants for the "redirect" option */
-#define REDIRECT_ON         1
-#define REDIRECT_OFF        2
-
 #if defined (USE_WIN32)
 #define ICON_IMAGE HICON
 #elif defined(__APPLE__)
@@ -73,7 +69,8 @@ typedef enum {
 typedef enum {
     LOG_ID_SEQUENTIAL,
     LOG_ID_UNIQUE,
-    LOG_ID_THREAD
+    LOG_ID_THREAD,
+    LOG_ID_PROCESS
 } LOG_ID;
 
 typedef enum {
@@ -99,8 +96,10 @@ typedef struct name_list_struct {
 } NAME_LIST;
 
 typedef struct sockaddr_list {                          /* list of addresses */
-    SOCKADDR_UNION *addr;                           /* the list of addresses */
-    unsigned *rr_ptr, rr_val;             /* current address for round-robin */
+    struct sockaddr_list *parent;   /* used by copies to locate their parent */
+    SOCKADDR_UNION *addr;                     /* array of resolved addresses */
+    SSL_SESSION **session;                /* array of cached client sessions */
+    unsigned rr;                          /* current address for round-robin */
     unsigned num;                             /* how many addresses are used */
     int passive;                                         /* listening socket */
     NAME_LIST *names;                          /* a list of unresolved names */
@@ -113,7 +112,7 @@ typedef enum {
 #endif /* !defined(OPENSSL_NO_COMP) */
 
 typedef struct {
-        /* some data for SSL initialization in ssl.c */
+        /* some data for TLS initialization in ssl.c */
 #ifndef OPENSSL_NO_COMP
     COMP_TYPE compression;                               /* compression type */
 #endif /* !defined(OPENSSL_NO_COMP) */
@@ -179,7 +178,7 @@ typedef struct psk_table_struct {
 
 typedef struct service_options_struct {
     struct service_options_struct *next;   /* next node in the services list */
-    SSL_CTX *ctx;                                            /*  SSL context */
+    SSL_CTX *ctx;                                            /*  TLS context */
     char *servname;        /* service name for logging & permission checking */
 
         /* service-specific data for stunnel.c */
@@ -382,7 +381,7 @@ typedef enum {
 
 typedef struct {
     jmp_buf err; /* 64-bit platforms require jmp_buf to be 16-byte aligned */
-    SSL *ssl; /* SSL connection */
+    SSL *ssl; /* TLS connection */
     SERVICE_OPTIONS *opt;
     TLS_DATA *tls;
 
@@ -390,6 +389,7 @@ typedef struct {
     socklen_t peer_addr_len;
     SOCKADDR_UNION *bind_addr; /* address to bind() the socket */
     SOCKADDR_LIST connect_addr; /* either copied or resolved dynamically */
+    unsigned idx; /* actually connected address in connect_addr */
     FD local_rfd, local_wfd; /* read and write local descriptors */
     FD remote_fd; /* remote file descriptor */
         /* IP for explicit local bind or transparent proxy */
@@ -400,13 +400,12 @@ typedef struct {
 
     /* data for transfer() function */
     char sock_buff[BUFFSIZE]; /* socket read buffer */
-    char ssl_buff[BUFFSIZE]; /* SSL read buffer */
+    char ssl_buff[BUFFSIZE]; /* TLS read buffer */
     size_t sock_ptr, ssl_ptr; /* index of the first unused byte */
     FD *sock_rfd, *sock_wfd; /* read and write socket descriptors */
-    FD *ssl_rfd, *ssl_wfd; /* read and write SSL descriptors */
-    uint64_t sock_bytes, ssl_bytes; /* bytes written to socket and SSL */
+    FD *ssl_rfd, *ssl_wfd; /* read and write TLS descriptors */
+    uint64_t sock_bytes, ssl_bytes; /* bytes written to socket and TLS */
     s_poll_set *fds; /* file descriptors */
-    uintptr_t redirect; /* redirect to another destination after failed auth */
 } CLI;
 
 /**************************************** prototypes for stunnel.c */
@@ -487,17 +486,15 @@ int cron_init(void);
 
 /**************************************** prototypes for ssl.c */
 
-extern int index_cli, index_opt, index_redirect, index_addr;
+extern int index_ssl_cli, index_ssl_ctx_opt;
+extern int index_session_authenticated, index_session_connect_address;
 
 int ssl_init(void);
 int ssl_configure(GLOBAL_OPTIONS *);
 
 /**************************************** prototypes for ctx.c */
 
-typedef struct {
-    SERVICE_OPTIONS *section;
-    char pass[PEM_BUFSIZE];
-} UI_DATA;
+extern SERVICE_OPTIONS *current_section;
 
 #ifndef OPENSSL_NO_DH
 extern DH *dh_params;
@@ -798,7 +795,7 @@ void ui_new_log(const char *);
 void message_box(LPCTSTR, const UINT);
 #endif /* USE_WIN32 */
 
-int passwd_cb(char *, int, int, void *);
+int ui_passwd_cb(char *, int, int, void *);
 #ifndef OPENSSL_NO_ENGINE
 UI_METHOD *UI_stunnel(void);
 #endif /* !defined(OPENSSL_NO_ENGINE) */
